/*
 * smartctl.cpp
 *
 * Home page of code is: https://www.smartmontools.org
 *
 * Copyright (C) 2002-11 Bruce Allen
 * Copyright (C) 2008-21 Christian Franke
 * Copyright (C) 2000 Michael Cornwell <cornwell@acm.org>
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

#include "config.h"
#define __STDC_FORMAT_MACROS 1 // enable PRI* for C++

#include <errno.h>
#include <inttypes.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdexcept>
#include <getopt.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#if defined(__FreeBSD__)
#include <sys/param.h>
#endif

#include "atacmds.h"
#include "dev_interface.h"
#include "ataprint.h"
#include "knowndrives.h"
#include "scsicmds.h"
#include "scsiprint.h"
#include "nvmeprint.h"
#include "smartctl.h"
#include "utility.h"

const char *smartctl_cpp_cvsid = "$Id: smartctl.cpp 5333 2022-02-26 00:15:22Z dpgilbert $" CONFIG_H_CVSID SMARTCTL_H_CVSID;

// Globals to control printing
bool printing_is_switchable = false;
bool printing_is_off = false;
std::string nvmeTemp = "";
std::string sataTemp = "";

// Control JSON output
json jglb;
static bool print_as_json = false;
static json::print_options print_as_json_options;
static bool print_as_json_output = false;
static bool print_as_json_impl = false;
static bool print_as_json_unimpl = false;

static void printslogan()
{
  jout("%s\n", format_version_info("smartctl").c_str());
}

static void UsageSummary()
{
  pout("\nUse smartctl -h to get a usage summary\n\n");
  return;
}

static void js_initialize(int argc, char **argv, bool verbose)
{
  if (jglb.is_enabled())
    return;
  jglb.enable();
  if (verbose)
    jglb.set_verbose();

  // Major.minor version of JSON format
  jglb["json_format_version"][0] = 1;
  jglb["json_format_version"][1] = 0;

  // Smartctl version info
  json::ref jref = jglb["smartctl"];
  int ver[3] = {0, 0, 0};
  sscanf(PACKAGE_VERSION, "%d.%d.%d", ver, ver + 1, ver + 2);
  jref["version"][0] = ver[0];
  jref["version"][1] = ver[1];
  if (ver[2] > 0)
    jref["version"][2] = ver[2];

#ifdef SMARTMONTOOLS_SVN_REV
  jref["svn_revision"] = SMARTMONTOOLS_SVN_REV;
#endif
  jref["platform_info"] = smi()->get_os_version_str();
#ifdef BUILD_INFO
  jref["build_info"] = BUILD_INFO;
#endif

  jref["argv"][0] = "smartctl";
  for (int i = 1; i < argc; i++)
    jref["argv"][i] = argv[i];
}

static std::string getvalidarglist(int opt);

// Values for  --long only options, see parse_options()
enum
{
  opt_identify = 1000,
  opt_scan,
  opt_scan_open,
  opt_set,
  opt_smart
};

/* Returns a string containing a formatted list of the valid arguments
   to the option opt or empty on failure. Note 'v' case different */
static std::string getvalidarglist(int opt)
{
  switch (opt)
  {
  case 'q':
    return "errorsonly, silent, noserial";
  case 'd':
    return smi()->get_valid_dev_types_str() + ", auto, test";
  case 'T':
    return "normal, conservative, permissive, verypermissive";
  case 'b':
    return "warn, exit, ignore";
  case 'B':
    return "[+]<FILE_NAME>";
  case 'r':
    return "ioctl[,N], ataioctl[,N], scsiioctl[,N], nvmeioctl[,N]";
  case opt_smart:
  case 'o':
  case 'S':
    return "on, off";
  case 'l':
    return "error, selftest, selective, directory[,g|s], "
           "xerror[,N][,error], xselftest[,N][,selftest], "
           "background, sasphy[,reset], sataphy[,reset], "
           "scttemp[sts,hist], scttempint,N[,p], "
           "scterc[,N,M][,p|reset], devstat[,N], defects[,N], "
           "ssd, gplog,N[,RANGE], smartlog,N[,RANGE], "
           "nvmelog,N,SIZE, tapedevstat, zdevstat, envrep";
  case 'P':
    return "use, ignore, show, showall";
  case 't':
    return "offline, short, long, conveyance, force, vendor,N, select,M-N, "
           "pending,N, afterselect,[on|off]";
  case 'F':
    return std::string(get_valid_firmwarebug_args()) + ", swapid";
  case 'n':
    return "never, sleep[,STATUS[,STATUS2]], standby[,STATUS[,STATUS2]], "
           "idle[,STATUS[,STATUS2]]";
  case 'f':
    return "old, brief, hex[,id|val]";
  case 'g':
    return "aam, apm, dsn, lookahead, security, wcache, rcache, wcreorder, wcache-sct";
  case opt_set:
    return "aam,[N|off], apm,[N|off], dsn,[on|off], lookahead,[on|off], security-freeze, "
           "standby,[N|off|now], wcache,[on|off], rcache,[on|off], wcreorder,[on|off[,p]], "
           "wcache-sct,[ata|on|off[,p]]";
  case 's':
    return getvalidarglist(opt_smart) + ", " + getvalidarglist(opt_set);
  case 'j':
    return "c, g, i, o, s, u, v, y";
  case opt_identify:
    return "n, wn, w, v, wv, wb";
  case 'v':
  default:
    return "";
  }
}

/* Prints the message "=======> VALID ARGUMENTS ARE: <LIST> \n", where
   <LIST> is the list of valid arguments for option opt. */
static void printvalidarglistmessage(int opt)
{
  if (opt == 'v')
  {
    jerr("=======> VALID ARGUMENTS ARE:\n\thelp\n%s\n<=======\n",
         create_vendor_attribute_arg_list().c_str());
  }
  else
  {
    // getvalidarglist() might produce a multiline or single line string.  We
    // need to figure out which to get the formatting right.
    std::string s = getvalidarglist(opt);
    char separator = strchr(s.c_str(), '\n') ? '\n' : ' ';
    jerr("=======> VALID ARGUMENTS ARE:%c%s%c<=======\n", separator, s.c_str(), separator);
  }

  return;
}

// Checksum error mode
enum checksum_err_mode_t
{
  CHECKSUM_ERR_WARN,
  CHECKSUM_ERR_EXIT,
  CHECKSUM_ERR_IGNORE
};

static checksum_err_mode_t checksum_err_mode = CHECKSUM_ERR_WARN;

static void scan_devices(const smart_devtype_list &types, bool with_open, char **argv);

// Printing functions

__attribute_format_printf(3, 0) static void vjpout(bool is_js_impl, const char *msg_severity,
                                                   const char *fmt, va_list ap)
{
  if (!print_as_json)
  {
    // Print out directly
    vprintf(fmt, ap);
    fflush(stdout);
  }
  else
  {
    // Add lines to JSON output
    static char buf[1024];
    static char *bufnext = buf;
    vsnprintf(bufnext, sizeof(buf) - (bufnext - buf), fmt, ap);
    for (char *p = buf, *q;; p = q)
    {
      if (!(q = strchr(p, '\n')))
      {
        // Keep remaining line for next call
        for (bufnext = buf; *p; bufnext++, p++)
          *bufnext = *p;
        break;
      }
      *q++ = 0; // '\n' -> '\0'

      static int lineno = 0;
      lineno++;
      if (print_as_json_output)
      {
        // Collect full output in array
        static int outindex = 0;
        jglb["smartctl"]["output"][outindex++] = p;
      }
      if (!*p)
        continue; // Skip empty line

      if (msg_severity)
      {
        // Collect non-empty messages in array
        static int errindex = 0;
        json::ref jref = jglb["smartctl"]["messages"][errindex++];
        jref["string"] = p;
        jref["severity"] = msg_severity;
      }

      if ((is_js_impl && print_as_json_impl) || (!is_js_impl && print_as_json_unimpl))
      {
        // Add (un)implemented non-empty lines to global object
        jglb[strprintf("smartctl_%04d_%c", lineno,
                       (is_js_impl ? 'i' : 'u'))
                 .c_str()] = p;
      }
    }
  }
}

// Default: print to stdout
// --json: ignore
// --json=o: append to "output" array
// --json=u: add "smartctl_NNNN_u" element(s)
void pout(const char *fmt, ...)
{
  if (printing_is_off)
    return;
  if (print_as_json && !(print_as_json_output || print_as_json_impl || print_as_json_unimpl))
    return;

  va_list ap;
  va_start(ap, fmt);
  vjpout(false, 0, fmt, ap);
  va_end(ap);
}

// Default: Print to stdout
// --json: ignore
// --json=o: append to "output" array
// --json=i: add "smartctl_NNNN_i" element(s)
void jout(const char *fmt, ...)
{
  if (printing_is_off)
    return;
  if (print_as_json && !(print_as_json_output || print_as_json_impl || print_as_json_unimpl))
    return;

  va_list ap;
  va_start(ap, fmt);
  vjpout(true, 0, fmt, ap);
  va_end(ap);
}

// Default: print to stdout
// --json: append to "messages"
// --json=o: append to "output" array
// --json=i: add "smartctl_NNNN_i" element(s)
void jinf(const char *fmt, ...)
{
  if (printing_is_off)
    return;

  va_list ap;
  va_start(ap, fmt);
  vjpout(true, "information", fmt, ap);
  va_end(ap);
}

void jwrn(const char *fmt, ...)
{
  if (printing_is_off)
    return;

  va_list ap;
  va_start(ap, fmt);
  vjpout(true, "warning", fmt, ap);
  va_end(ap);
}

void jerr(const char *fmt, ...)
{
  if (printing_is_off)
    return;

  va_list ap;
  va_start(ap, fmt);
  vjpout(true, "error", fmt, ap);
  va_end(ap);
}

static char startup_datetime_buf[DATEANDEPOCHLEN];

// Print smartctl start-up date and time and timezone
void jout_startup_datetime(const char *prefix)
{
  jout("%s%s\n", prefix, startup_datetime_buf);
}

// Globals to set failuretest() policy
bool failuretest_conservative = false;
unsigned char failuretest_permissive = 0;

// Compares failure type to policy in effect, and either exits or
// simply returns to the calling routine.
// Used in ataprint.cpp and scsiprint.cpp.
void failuretest(failure_type type, int returnvalue)
{
  // If this is an error in an "optional" SMART command
  if (type == OPTIONAL_CMD)
  {
    if (!failuretest_conservative)
      return;
    pout("An optional SMART command failed: exiting. Remove '-T conservative' option to continue.\n");
    throw int(returnvalue);
  }

  // If this is an error in a "mandatory" SMART command
  if (type == MANDATORY_CMD)
  {
    if (failuretest_permissive--)
      return;
    pout("A mandatory SMART command failed: exiting. To continue, add one or more '-T permissive' options.\n");
    throw int(returnvalue);
  }

  throw std::logic_error("failuretest: Unknown type");
}

// Used to warn users about invalid checksums. Called from atacmds.cpp.
// Action to be taken may be altered by the user.
void checksumwarning(const char *string)
{
  // user has asked us to ignore checksum errors
  if (checksum_err_mode == CHECKSUM_ERR_IGNORE)
    return;

  pout("Warning! %s error: invalid SMART checksum.\n", string);

  // user has asked us to fail on checksum errors
  if (checksum_err_mode == CHECKSUM_ERR_EXIT)
    throw int(FAILSMART);
}

// Return info string about device protocol
static const char *get_protocol_info(const smart_device *dev)
{
  switch ((int)dev->is_ata() | ((int)dev->is_scsi() << 1) | ((int)dev->is_nvme() << 2))
  {
  case 0x1:
    return "ATA";
  case 0x2:
    return "SCSI";
  case 0x3:
    return "ATA+SCSI";
  case 0x4:
    return "NVMe";
  default:
    return "Unknown";
  }
}

// Add JSON device info
static void js_device_info(const json::ref &jref, const smart_device *dev)
{
  jref["name"] = dev->get_dev_name();
  jref["info_name"] = dev->get_info_name();
  jref["type"] = dev->get_dev_type();
  jref["protocol"] = get_protocol_info(dev);
}

// Device scan
// smartctl [-d type] --scan[-open] -- [PATTERN] [smartd directive ...]
void scan_devices(const smart_devtype_list &types, bool with_open, char **argv)
{
  bool dont_print = !(ata_debugmode || scsi_debugmode || nvme_debugmode);

  const char *pattern = 0;
  int ai = 0;
  if (argv[ai] && argv[ai][0] != '-')
    pattern = argv[ai++];

  smart_device_list devlist;
  printing_is_off = dont_print;
  bool ok = smi()->scan_smart_devices(devlist, types, pattern);
  printing_is_off = false;

  if (!ok)
  {
    pout("# scan_smart_devices: %s\n", smi()->get_errmsg());
    return;
  }

  for (unsigned i = 0; i < devlist.size(); i++)
  {
    smart_device_auto_ptr dev(devlist.release(i));
    json::ref jref = jglb["devices"][i];

    if (with_open)
    {
      printing_is_off = dont_print;
      dev.replace(dev->autodetect_open());
      printing_is_off = false;
    }

    js_device_info(jref, dev.get());

    if (with_open && !dev->is_open())
    {
      jout("# %s -d %s # %s, %s device open failed: %s\n", dev->get_dev_name(),
           dev->get_dev_type(), dev->get_info_name(),
           get_protocol_info(dev.get()), dev->get_errmsg());
      jref["open_error"] = dev->get_errmsg();
      continue;
    }

    jout("%s -d %s", dev->get_dev_name(), dev->get_dev_type());
    if (!argv[ai])
      jout(" # %s, %s device\n", dev->get_info_name(), get_protocol_info(dev.get()));
    else
    {
      for (int j = ai; argv[j]; j++)
        jout(" %s", argv[j]);
      jout("\n");
    }

    if (dev->is_open())
      dev->close();
  }
}

// Main program without exception handling
int main_worker(const char *name)
{
  nvmeTemp.clear();
  sataTemp.clear();
  // Throw if runtime environment does not match compile time test.
  check_config();

  // Initialize interface
  smart_interface::init();
  if (!smi())
    return 1;

  // Parse input arguments
  const char *type = 0;
  ata_print_options ataopts;
  scsi_print_options scsiopts;
  nvme_print_options nvmeopts;
  bool print_type_only = false;
  {
    ataopts.smart_vendor_attrib = scsiopts.smart_vendor_attrib = nvmeopts.smart_vendor_attrib = true;
  }

  // Store formatted current time for jout_startup_datetime()
  // Output as JSON regardless of '-i' option
  {
    time_t now = time(nullptr);
    dateandtimezoneepoch(startup_datetime_buf, now);
    jglb["local_time"] += {{"time_t", now}, {"asctime", startup_datetime_buf}};
  }

  smart_device_auto_ptr dev;
  if (!strcmp(name, "-"))
  {
    // Parse "smartctl -r ataioctl,2 ..." output from stdin
    if (type || print_type_only)
    {
      pout("-d option is not allowed in conjunction with device name \"-\".\n");
      UsageSummary();
      return FAILCMD;
    }
    dev = get_parsed_ata_device(smi(), name);
  }
  else
    // get device of appropriate type
    dev = smi()->get_smart_device(name, type);
  if (!dev)
  {
    jerr("%s: %s\n", name, smi()->get_errmsg());
    if (type)
      printvalidarglistmessage('d');
    else
      pout("Please specify device type with the -d option.\n");
    UsageSummary();
    return FAILCMD;
  }

  if (print_type_only)
    // Report result of first autodetection
    pout("%s: Device of type '%s' [%s] detected\n",
         dev->get_info_name(), dev->get_dev_type(), get_protocol_info(dev.get()));

  if (dev->is_ata() && ataopts.powermode >= 2 && dev->is_powered_down())
  {
    jinf("Device is in STANDBY (OS) mode, exit(%d)\n", ataopts.powerexit);
    return ataopts.powerexit;
  }

  // Open device
  {
    // Save old info
    smart_device::device_info oldinfo = dev->get_info();

    // Open with autodetect support, may return 'better' device
    dev.replace(dev->autodetect_open());

    // Report if type has changed
    if ((ata_debugmode || scsi_debugmode || nvme_debugmode || print_type_only) && oldinfo.dev_type != dev->get_dev_type())
      pout("%s: Device open changed type from '%s' to '%s'\n",
           dev->get_info_name(), oldinfo.dev_type.c_str(), dev->get_dev_type());
  }
  if (!dev->is_open())
  {
    jerr("Smartctl open device: %s failed: %s\n", dev->get_info_name(), dev->get_errmsg());
    return FAILDEV;
  }

  // Add JSON info similar to --scan output
  js_device_info(jglb["device"], dev.get());

  // now call appropriate ATA or SCSI routine
  int retval = 0;
  if (print_type_only)
    jout("%s: Device of type '%s' [%s] opened\n",
         dev->get_info_name(), dev->get_dev_type(), get_protocol_info(dev.get()));
  else if (dev->is_ata())
    retval = ataPrintMain(dev->to_ata(), ataopts);
  else if (dev->is_scsi())
    retval = scsiPrintMain(dev->to_scsi(), scsiopts);
  else if (dev->is_nvme())
    retval = nvmePrintMain(dev->to_nvme(), nvmeopts);
  else
    // we should never fall into this branch!
    pout("%s: Neither ATA, SCSI nor NVMe device\n", dev->get_info_name());

  dev->close();
  return retval;
}
